#lang scheme

(define table
  `([main [("apple") ,(lambda () 'apple)]
          [("banana") ,(lambda () 'banana)]
          [(fruit-salad "!") ,(lambda (x)
                                `(,x !))]]
    [fruit-salad [(fruit fruit) ,(lambda (x y)
                                   (list x y))]
                 [(fruit fruit fruit) ,(lambda (x y z)
                                         (list x y z))]]
    [fruit [("orange") ,(lambda ()'orange)]
           [("grape") ,(lambda () 'grape)]]))

(define (parse what table stream)
  (local [(define defn (assq what table))]
    (if defn
        (parse-alts (rest defn) table stream)
        (error 'parse "no definition for ~a" what))))

(define (parse-alts alts table stream)
  (if (null? alts)
      false
      (or (parse-case (first alts) table stream)
          (parse-alts (rest alts) table stream))))

(define (parse-case case table stream)
  (local [(define seq (first case))
          (define handler (second case))]
    (local [(define parsed+stream (parse-sequence seq table stream))]
      (and parsed+stream
           (cons (apply handler (car parsed+stream))
                 (cdr parsed+stream))))))

(define (parse-sequence seq table stream)
  (cond
    [(empty? seq) (cons empty stream)]
    [else (local [(define first-parsed+stream
                    (parse-item (first seq) table stream))]
            (and first-parsed+stream
                 (local [(define rest-parsed+stream
                           (parse-sequence (rest seq) table 
                                           (cdr first-parsed+stream)))]
                   (and rest-parsed+stream
                        (cons (append (car first-parsed+stream)
                                      (car rest-parsed+stream))
                              (cdr rest-parsed+stream))))))]))

(define (parse-item item table stream)
  (cond
    [(string? item)
     (if (and (cons? stream)
              (equal? item (first stream)))
         (cons empty (cdr stream))
         false)]
    [else
     (let ([x+stream (parse item table stream)])
       (and x+stream 
            (cons (list (car x+stream))
                  (cdr x+stream))))]))

(parse 'main table '("orange" "orange" "!"))
(parse 'main table '("orange" "orange" "orange" "!"))
