#lang scheme

(define table
  `([main [("apple") ,(lambda () 'apple)]
          [("banana") ,(lambda () 'banana)]
          [(fruit-salad "grape" "!") ,(lambda (x)
                                        `(,x !))]]
    [fruit-salad [(fruit fruit) ,(lambda (x y)
                                   (list x y))]
                 [(fruit fruit fruit) ,(lambda (x y z)
                                         (list x y z))]]
    [fruit [("orange") ,(lambda () 'orange)]
           [("grape") ,(lambda () 'grape)]]))

(define (parse what table stream 
               success-k fail-k)
  (local [(define defn (assq what table))]
    (if defn
        (parse-alts (rest defn) table stream 
                    success-k fail-k)
        (error 'parse "no definition for ~a" what))))

(define (parse-alts alts table stream success-k fail-k)
  (if (null? alts)
      (fail-k)
      (parse-case (first alts) table stream
                  success-k
                  (lambda ()
                    (parse-alts (rest alts) table stream
                                success-k fail-k)))))

(define (parse-case case table stream success-k fail-k)
  (local [(define seq (first case))
          (define handler (second case))]
    (parse-sequence seq table stream
                    (lambda (vals stream)
                      (success-k (apply handler vals) stream))
                    fail-k)))

(define (parse-sequence seq table stream success-k fail-k)
  (cond
    [(empty? seq) (success-k empty stream)]
    [else 
     (parse-item (first seq) table stream
                 (lambda (fst-vals stream)
                   (parse-sequence (rest seq) table stream
                                   (lambda (snd-vals stream)
                                     (success-k (append fst-vals
                                                        snd-vals)
                                                stream))
                                   fail-k))
                 fail-k)]))

(define (parse-item item table stream success-k fail-k)
  (cond
    [(string? item)
     (if (and (cons? stream)
              (equal? item (first stream)))
         (success-k empty (rest stream))
         (fail-k))]
    [else
     (parse item table stream
            (lambda (v stream)
              (success-k (list v) stream))
            fail-k)]))

(parse 'fruit-salad table '("orange" "orange" "!")
       (lambda (val stream) val)
       (lambda () 'nope!))

(parse 'main table '("orange" "orange" "orange" "!")
       list
       (lambda () 'nope!))
